/* Version 81m - for Atmel 328 - used ARDUINO VERSION 0020, version 0011 doesn't support the 328 part  */
//GPL Software    
// m-version - Mike Luckham, Sunrise Systems Software Email sunrise@sunsys.net

// when making a new release, update the display version number by finding this code:
//    LCD::print(getStr(PSTR("  MPGuino v0.81m")));  


#include <avr/pgmspace.h>  
#include <Wire.h>

// UNCOMMENT THIS LINE TO USE DEFAULT VALUES IN parms[] INSTEAD OF READING THE VALUES FROM EEPROM IN load()
//#define usedefaults true

// CHANGES FOR .82m
//  - added BIG display of INSTANT L/HK, to better see cause and effect when using the throttle, going uphill/downhill, winds, etc
//  - added I2C data output
//  - when compiled with IDE version 020, binary sketch is 21490 bytes (out of a 30720 byte maximum)
//
// CHANGES FOR .81m
//  - added SKelly font hack from http://ecomodder.com/wiki/index.php/Code_hacks (rounds the corners and makes the fonts easier to read)
//  - added big RPM and vehicle speed displays
//  - when compiled with IDE version 020, binary sketch is 19362 bytes (out of a 30720 byte maximum)
//
// CHANGES FOR .80m
//  - bug fixes for new tank/distance calibration method
//  - added alternating L/100KM / MPG display for Tank and Current, it works but displays 999 MPG when Current L/100KM is zero
//  - modified edit parameter to leave cursor under OK, instead of the first non-zero digit
//  - don't calculate MaxLoop (for CPU % statistic) while the user is operating the configuration menu
//  - when compiled with IDE version 013, binary sketch is 19356 bytes (out of a 30720 byte maximum)
//
// CHANGES FOR .79m
//  - corrected display of L/HK on Tank/Current tDisplay() function (upper left - was displaying average speed, want to show fuel consumption in L/HK) and Custom (lower left)
//  - new 'simpler calibration' method introduced in 0.77m does not work

// CHANGES BASED ON 0.76m, BECOMES 0.77m
//  - added Tank save to EEPROM, reload on power up
//  - add two Setup parameters for simpler calibration of fuel, distance - just correct the 'Tank Km' and 'Tank Litres' values shown, the code does the rest!
//  - added comments to the code

// MEMORY ON AN ATMEL168 IS TRULY TIGHT - RUNTIME SWITCH BETWEEN UNITS IS TBD - OR UPGRADE TO ATMEL328 PART

#define metric
#define i2cslave

// SKelly's FONT HACK
#define CFG_BIGFONT_TYPE  2  // 1=Original 2=Modified (adds about 20 bytes to code size)

// metric conversion factors
#define GALLONS2LITRES    3.78541178
#define MILES2KILOMETRES  1.609344
#define POUNDS2KILOGRAMS  0.45359237

// conversion factor from L/100KM to MPG (US) 235.214583 ... multiplied by 1000000 for correct conversion using integer math
#define L100KM2USMPG  235214583

// I2C slave address
#define I2CADDRESS 42

// there must be the same number of elements in parms[] and parmLabels[]

#ifdef metric

//unsigned long  parms[]={95ul,5100ul,132086026ul,3ul,420000000ul,10300ul,500ul,1088ul,0ul,2ul,1ul};//default values
unsigned long  parms[]={6ul,9926ul,57442679ul,1ul,420000000ul,42000ul,500ul,1181ul,0ul,2ul,1ul,0,0};//values for 1995 SOHC 2-litre Neon Sport
//unsigned long  parms[]={6ul,5100ul,73766100ul,1ul,420000000ul,42000ul,500ul,1181ul,0ul,2ul,1ul,0,0};//values for Civic VS 92


char *  parmLabels[]={"Contrast","VSS Pulses/Km", "MicroSec/Litre","Pulses/2 revs","Timout(microSec)","Tank Litr * 1000","Injector DelayuS","Weight (kgs)","Scratchpad(odo?)","VSS Delay ms","0-USA 1-METRIC","Tank Km * 1000","Tank Litres*1000"};

#define DISTANCECHAR 'K'
#define DISTANCE2CHAR "KM"
#define FUELCHAR 'L'
#define FUEL2CHAR "LI"
#define EFFICIENCYSTRING "KM/L"
#define EFFICIENCY2CHAR "KL"
#define CONSUMPTIONSTRING "L/KM"
#define COMSUMPTION2CHAR "LK"
#define CONSUMPTION100STRING "L/HK"
#define CONSUMPTIONRATESTRING "L/HR"
#define C100STRING "HK"
#define C100CHAR 'H'
#define HOURCHAR 'h'
#define SPEED2CHAR "KH"
#define SPEED4CHAR "KPH "
#define RPM4CHAR "RPM "

#else

unsigned long  parms[]={95ul,8208ul,500000000ul,3ul,420000000ul,10300ul,500ul,2400ul,0ul,2ul,0,0,0};//default values
char *  parmLabels[]={"Contrast","VSS Pulses/Mile", "MicroSec/Gallon","Pulses/2 revs","Timout(microSec)","Tank Gal * 1000","Injector DelayuS","Weight (lbs)","Scratchpad(odo?)","VSS Delay ms","0-USA 1-METRIC","Tank Miles*1000","Tank Gals*1000"};
#define DISTANCECHAR 'M'
#define DISTANCE2CHAR "MI"
#define FUELCHAR 'G'
#define FUEL2CHAR "GA"
#define EFFICIENCYSTRING "MPG "
#define EFFICIENCY2CHAR "MG"
#define CONSUMPTIONSTRING "G/M"
#define CONSUMPTION100STRING "G/HM"
#define COMSUMPTION2CHAR "GM"
#define SPEED2CHAR "MH"
#define SPEED4CHAR "MPH "
#define RPM4CHAR "RPM "

#endif

// CONFIGURATION PARAMETERS STORED IN EEPROM
// CHANGING ANY VALUE MARKED (*) ALTERS DISPLAY OF FUEL AND/OR DISTANCE UNITS
//
//   contrastIdx                  - LCD contrast adjustment
//   vssPulsesPerDistanceUnitIdx  * used for speed, distance and fuel-distance calculations
//   microSecondsPerFuelUnitIdx   * used for fuel consumption and fuel-distance calculations
//   injPulsesPer2Revolutions     - used only for RPM calculation
//   currentTripResetTimeoutUSIdx - vehicle-stopped-engine-off interval to decide when a Trip is finished
//   tankSizeIdx                  - not used in the code
//   injectorSettleTimeIdx        * should be left as the default value
//   weightIdx                    - not used in the code
//   scratchpadIdx                - not used in the code
//   vsspause                     * should be left as the default value
//   units                        - 0=USA  1=METRIC/CANADA
//   calibrationDistanceIdx       * used to correct distance calibration
//   calibrationFuelIdx           * used to correct fuel calibration

#define contrastIdx 0  //do contrast first to get display dialed in
#define vssPulsesPerDistanceUnitIdx 1
#define microSecondsPerFuelUnitIdx 2
#define injPulsesPer2Revolutions 3
#define currentTripResetTimeoutUSIdx 4
#define tankSizeIdx 5 
#define injectorSettleTimeIdx 6
#define weightIdx 7
#define scratchpadIdx 8
#define vsspause 9
#define units 10
#define calibrationDistanceIdx 11
#define calibrationFuelIdx 12
#define parmsLength (sizeof(parms)/sizeof(unsigned long)) //array size      

// CALIBRATION USING THE MANUAL METHOD
//     Wait until you will go on a trip of at least an hour's driving
//     Grab a calculator, a pen, and paper
//     Go to the gas station and top up the tank
//     Reset your car's odometer or trip meter
//     And do a Tank Reset on the MPGuino
//     Drive - the farther and longer, the better your accuracy will be
//     Go back to the gas station and top up the tank again
//     Write down the amount of fuel added
//     Write down your odometer reading to get the actual distance travelled
//     In the gas station parking lot:
//       Navigate to the MPGuino Tank screen
//       Write down the fuel used and the distance the MPGuino says you travelled
//     Go into MPGuino Setup to adjust the two critical values:
//        write down the current value for VSS Pulses/Km
//        calculate the new value:  new value = current value * MPGguino-distance / Odometer-distance
//        enter the new value into the Setup screen
//        write down the current value for MicroSec/Litre
//        calculate the new value: new value = current value * MPGuino-fuel / GasStationPump-fuel
//        enter the new value into the Setup screen
//     Go through the rest of the Setup to save the results

// CALIBRATION USING THE SIMPLER METHOD:
//     Go to the gas station and top up the tank
//     Reset your car's odometer or trip meter
//     And do a Tank Reset on the MPGuino
//     Drive - the farther and longer, the better your accuracy will be
//     Go back to the gas station and top up the tank again
//     Write down the amount of fuel added
//     Write down your odometer reading to get the actual distance travelled
//     In the gas station parking lot:
//     Go into MPGuino Setup to adjust the two calibration values:
//        go to "Tank Km * 1000", it contains the current Tank Distance travelled
//        multiply your Trip Meter x 1000 and input this number instead
//        go to "Tank Litres*1000", it contains the current Tank Fuel used
//        multiply the amount of fuel added to the tank x 1000 and input this number instead
//     Go through the rest of the Setup to save the results


// EEPROM - ATmega128 has 512 bytes, ATmega328 as 1024 bytes

// Bytes 000-099   Reserved for Configuration
//       100-199   Reserved for Current Trip
//       200-299   Reserved for Tank Trip
#define EEPROMOFFSET_CONFIGURATION 0
#define EEPROMOFFSET_CURRENTTRIP 100
#define EEPROMOFFSET_TANKTRIP 200


#define nil 3999999999ul
 
// change this signature value if parameter list is changed, it causes the current
// values in EEPROM to be overwritten with the default values in parms[]

#define guinosigpremetric B10100111 
#define guinosigpresaveTank B11101000 
#define guinosig B11101001
#include <EEPROM.h>

//Vehicle Interface Pins      
#define InjectorOpenPin 2      
#define InjectorClosedPin 3      
#define VSSPin 14 //analog 0      
 
//LCD Pins      
#define DIPin 4 // register select RS      
#define DB4Pin 7       
#define DB5Pin 8       
#define DB6Pin 12       
#define DB7Pin 13      
#define ContrastPin 6      
#define EnablePin 5       
#define BrightnessPin 9      
 
//Button Pins
#define lbuttonPin 17 // Left Button, on analog 3,        
#define mbuttonPin 18 // Middle Button, on analog 4       
#define rbuttonPin 19 // Right Button, on analog 5       
 
#define vssBit 1     //  pin14 is a bitmask 1 on port C        
#define lbuttonBit 8 //  pin17 is a bitmask 8 on port C        
#define mbuttonBit 16 // pin18 is a bitmask 16 on port C        
#define rbuttonBit 32 // pin19 is a bitmask 32 on port C        
#define loopsPerSecond 2 // how many times will we try and loop in a second     


byte brightness[]={255,214,171,128}; //middle button cycles through these brightness settings      
#define brightnessLength (sizeof(brightness)/sizeof(byte)) //array size      
byte brightnessIdx=1;      


typedef void (* pFunc)(void);//type for display function pointers      

volatile unsigned long timer2_overflow_count;

/*** Set up the Events ***
We have our own ISR for timer2 which gets called about once a millisecond.
So we define certain event functions that we can schedule by calling addEvent
with the event ID and the number of milliseconds to wait before calling the event. 
The milliseconds is approximate.

Keep the event functions SMALL!!!  This is an interrupt!

*/
//event functions

void enableLButton(){PCMSK1 |= (1 << PCINT11);}
void enableMButton(){PCMSK1 |= (1 << PCINT12);}
void enableRButton(){PCMSK1 |= (1 << PCINT13);}
//array of the event functions
pFunc eventFuncs[] ={enableVSS, enableLButton,enableMButton,enableRButton};
#define eventFuncSize (sizeof(eventFuncs)/sizeof(pFunc)) 
//define the event IDs
#define enableVSSID 0
#define enableLButtonID 1
#define enableMButtonID 2
#define enableRButtonID 3
//ms counters
unsigned int eventFuncCounts[eventFuncSize];

//schedule an event to occur ms milliseconds from now
void addEvent(byte eventID, unsigned int ms){
  if( ms == 0)
    eventFuncs[eventID]();
  else
    eventFuncCounts[eventID]=ms;
}

/* this ISR gets called every 1.024 milliseconds, we will call that a millisecond for our purposes
go through all the event counts, 
  if any are non zero subtract 1 and call the associated function if it just turned zero.  */
ISR(TIMER2_OVF_vect){
  timer2_overflow_count++;
  for(byte eventID = 0; eventID < eventFuncSize; eventID++){
    if(eventFuncCounts[eventID]!= 0){
      eventFuncCounts[eventID]--;
      if(eventFuncCounts[eventID] == 0)
          eventFuncs[eventID](); 
    }  
  }
}

//for math calculations
unsigned long tmp1[2];
unsigned long tmp2[2];
unsigned long tmp3[2];

 
 
unsigned long maxLoopLength = 0; //see if we are overutilizing the CPU      
 
 
#define buttonsUp   lbuttonBit + mbuttonBit + rbuttonBit  // start with the buttons in the right state      
byte buttonState = buttonsUp;      
 
 
//overflow counter used by millis2()      
unsigned long lastMicroSeconds=millis2() * 1000;   
unsigned long microSeconds (void){     
  unsigned long tmp_timer2_overflow_count;    
  unsigned long tmp;    
  byte tmp_tcnt2;    
  cli(); //disable interrupts    
  tmp_timer2_overflow_count = timer2_overflow_count;    
  tmp_tcnt2 = TCNT2;    
  sei(); // enable interrupts    
  tmp = ((tmp_timer2_overflow_count << 8) + tmp_tcnt2) * 4;     
  if((tmp<=lastMicroSeconds) && (lastMicroSeconds<4290560000ul))    
    return microSeconds();     
  lastMicroSeconds=tmp;   
  return tmp;     
}    
 
unsigned long elapsedMicroseconds(unsigned long startMicroSeconds, unsigned long currentMicroseconds ){      
  if(currentMicroseconds >= startMicroSeconds)      
    return currentMicroseconds-startMicroSeconds;      
  return 4294967295 - (startMicroSeconds-currentMicroseconds);      
}      

unsigned long elapsedMicroseconds(unsigned long startMicroSeconds ){      
  return elapsedMicroseconds(startMicroSeconds, microSeconds());
}      
 
//Trip prototype      
class Trip{      
public:      
  unsigned long loopCount; //how long has this trip been running      
  unsigned long injPulses; //rpm      
  unsigned long injHiSec;// seconds the injector has been open      
  unsigned long injHius;// microseconds, fractional part of the injectors open       
  unsigned long injIdleHiSec;// seconds the injector has been open      
  unsigned long injIdleHius;// microseconds, fractional part of the injectors open       
  unsigned long vssPulses;//from the speedo      
  unsigned long vssECOPulses;//from the speedo      
  unsigned long vssPulseLength; // only used by instant
  //these functions actually return in thousandths      
  unsigned long distance();              // miles, kilometers
  unsigned long fuel();                    // gallons, litres
  unsigned long distanceperfuelunit();        // mpg, km/l
  unsigned long fuelperdistanceunit(unsigned long multiplier);        // gal/mi, l/km
  unsigned long vehiclespeed();                // mph, km/h
  unsigned long time(); //mmm.ss        
  unsigned long ecoDistance();  //how many "free" miles/km?        
  unsigned long idleFuel();  //how many gallons/litres spent standing still?        
  void update(Trip t);      
  void reset();    
  void loadfromEEPROM(unsigned int eepromoffset );
  void savetoEEPROM(unsigned int eepromoffset );
  Trip();      
};      
 
//LCD prototype      
namespace LCD{      
  void gotoXY(byte x, byte y);      
  void print(char * string);      
  void init();      
  void tickleEnable();      
  void cmdWriteSet();      
  void LcdCommandWrite(byte value);      
  void LcdDataWrite(byte value);      
  byte pushNibble(byte value);      
};      
 
 
//main objects we will be working with:      
unsigned long injHiStart; //for timing injector pulses      
Trip tmpTrip;      
Trip instant;      
Trip current;      
Trip tank;      


unsigned volatile long instInjStart=nil; 
unsigned volatile long tmpInstInjStart=nil; 
unsigned volatile long instInjEnd; 
unsigned volatile long tmpInstInjEnd; 
unsigned volatile long instInjTot; 
unsigned volatile long tmpInstInjTot;     
unsigned volatile long instInjCount; 
unsigned volatile long tmpInstInjCount;     


void processInjOpen(void){      
  injHiStart = microSeconds();  
}      
 
void processInjClosed(void){      
  long t =  microSeconds();
  long x = elapsedMicroseconds(injHiStart, t)- parms[injectorSettleTimeIdx];       
  if(x >0)
    tmpTrip.injHius += x;       
  tmpTrip.injPulses++;      

  if (tmpInstInjStart != nil) {
    if(x >0)
      tmpInstInjTot += x;     
    tmpInstInjCount++;
  } else {
    tmpInstInjStart = t;
  }
  
  tmpInstInjEnd = t;
}


volatile boolean vssFlop = 0;

void enableVSS(){
//    tmpTrip.vssPulses++; 
    vssFlop = !vssFlop;
}

unsigned volatile long lastVSS1;
unsigned volatile long lastVSSTime;
unsigned volatile long lastVSS2;

volatile boolean lastVssFlop = vssFlop;

//attach the vss/buttons interrupt      
ISR( PCINT1_vect ){   
  static byte vsspinstate=0;      
  byte p = PINC;//bypassing digitalRead for interrupt performance      
  if ((p & vssBit) != (vsspinstate & vssBit)){      
    addEvent(enableVSSID,parms[vsspause] ); //check back in a couple milli
  }
  if(lastVssFlop != vssFlop){
    lastVSS1=lastVSS2;
    unsigned long t = microSeconds();
    lastVSS2=elapsedMicroseconds(lastVSSTime,t);
    lastVSSTime=t;
    tmpTrip.vssPulses++; 
    tmpTrip.vssPulseLength += lastVSS2;
    lastVssFlop = vssFlop;
  }
  vsspinstate = p;      
  buttonState &= p;      
}       
 
 
 
#ifdef i2cslave

// I2C Slave event handler
//   I2C MASTER      I2C SLAVE (US)
//   request #    -->
//   selector     -->
//                 [Wire.OnReceive handler receives request #, selector]
//   request-data -->
//                 [Wire.OnRequest handler sends the unsigned long data value, little-endian style]
//               <-- byte 0 (bits 7-0)
//               <-- byte 1 (bits 15-8)
//               <-- byte 2 (bits 23-16)
//               <-- byte 3 (bits 31-24)

int reqnum = 0;
int selector = 0;
unsigned long datavalue = 0;

// receives the request # and selector from the I2C Master, whose code looks like this:
// ...
//    Wire.beginTransmission(ADDRESS);
//    Wire.send(requestnum);
//    Wire.send(selector);
//    Wire.endTransmission();
// ...
void receiveI2CEvent(int howMany) {
  while (Wire.available() > 0){
    reqnum = Wire.receive();      // what is requested
    selector = Wire.receive();    // 0=instant; 1=current; 2=tank
  }
  datavalue = -1;               // response for 'undefined'
} 

// transmits the requested data to the I2C Master, whose code looks like this:
// ...
//    Wire.requestFrom(ADDRESS,4);
//    while (Wire.available()) {
//      responsebyte = Wire.receive();
//    }
// ...
void sendI2Cdata() {
    switch (reqnum)
    {
      // units request
      case 0:  
#ifdef metric
               datavalue = 0;  // metric km, litres
#else
               datavalue = 1;  // US miles, gallons
#endif
               break;
      // speed request - responds with KPH or MPH in thousandths
      case 1:  datavalue = instantvehiclespeed();        
               break;
      // RPM request - responds with current RPM
      case 2:  datavalue = instantrpm();             
               break;
      // distance request - responds with distance travelled in Km or Miles in thousandths
      case 3:  switch (selector)
               {
                 case 0:  datavalue = instant.distance(); break;
                 case 1:  datavalue = current.distance(); break;
                 case 2:  datavalue = tank.distance(); break;
                 break;
               }
               break;
      // fuel consumption request - responds with Litres or Gallons in thousandths
      case 4:  switch (selector)
               {
                 case 0:  datavalue = instant.fuel(); break;
                 case 1:  datavalue = current.fuel(); break;
                 case 2:  datavalue = tank.fuel(); break;
                 break;
               }
               break;
      // fuel economy request - responds with Km/litre or Miles/Gallon in thousandths
      case 5:  switch (selector)
               {
                 case 0:  datavalue = instantdistanceperfuelunit(); break;
                 case 1:  datavalue = current.distanceperfuelunit(); break;
                 case 2:  datavalue = tank.distanceperfuelunit(); break;
                 break;
               }
               break;
      // fuel rate request - responds with L/HK or Gallons/Mile in thousandths
      case 6:  switch (selector)
               {
                 case 0:  datavalue = instantfuelperdistanceunit(100); break;
                 case 1:  datavalue = current.fuelperdistanceunit(100); break;
                 case 2:  datavalue = tank.fuelperdistanceunit(100); break;
                 break;
               }
               break;
      // ECOdistance request - responds with Km or Miles coasted in thousandths
      case 7:  switch (selector)
               {
                 case 0:  datavalue = instant.ecoDistance(); break;
                 case 1:  datavalue = current.ecoDistance(); break;
                 case 2:  datavalue = tank.ecoDistance(); break;
                 break;
               }
               break;
      // Idle Fuel consumption request - responds with Litres or Gallons used while idling in thousandths
      case 9:  switch (selector)
               {
                 case 0:  datavalue = current.idleFuel(); break;  // replace with instant fuel flow rate (Litres or Gallons per hour or per minute)
                 case 1:  datavalue = current.idleFuel(); break;
                 case 2:  datavalue = tank.idleFuel(); break;
                 break;
               }
               break;
      // Elapsed Time request
      case 10:  switch (selector)
               {
                 case 0:  datavalue = instant.time(); break;
                 case 1:  datavalue = current.time(); break;
                 case 2:  datavalue = tank.time(); break;
                 break;
               }
               break;
      // System Information request
      case 11:  switch (selector)
               {
                 case 0:  datavalue = 82; break;                 // version
                 case 1:  datavalue = 0; break;                  // reserved
                 case 2:  datavalue = memoryTest(); break;       // free memory in bytes
                 break;
               }
               break;
  
      default: // ignore it
               datavalue = -1;
               break;
    }
    
    // always replies on the bus with a four-byte long, sent little-endian
    Wire.send((uint8_t *)&datavalue,4);
}


#endif


pFunc displayFuncs[] ={ 
  doDisplayCustom, 
  doDisplayInstantCurrent, 
  doDisplayInstantTank, 
  doDisplayBigInstant, 
  doDisplayBigInstantFuelPerDistanceUnit,
  doDisplayBigVSS,
  doDisplayBigCurrent, 
  doDisplayBigCurrentFuelPerDistanceUnit,
  doDisplayBigTank,
  doDisplayBigTankFuelPerDistanceUnit,
  doDisplayCurrentTripData, 
  doDisplayTankTripData, 
  doDisplayECOIdleData, 
  doDisplaySystemInfo,
};      
#define displayFuncSize (sizeof(displayFuncs)/sizeof(pFunc)) //array size      
prog_char  * displayFuncNames[displayFuncSize]; 
byte newRun = 0;
void setup (void){
  init2();

  newRun = load();//load the default parameters
  byte x = 0;
  displayFuncNames[x++]=  PSTR("Custom  "); 
  displayFuncNames[x++]=  PSTR("Instant/Current "); 
  displayFuncNames[x++]=  PSTR("Instant/Tank "); 
  displayFuncNames[x++]=  PSTR("BIG Instant ");     // fuel economy distance/fuel unit
  displayFuncNames[x++]=  PSTR("BIG Instant ");     // fuel/100 distance units
  displayFuncNames[x++]=  PSTR("BIG VSS     ");     // vehicle speed
  displayFuncNames[x++]=  PSTR("BIG Current ");     // fuel economy distance/fuel unit
  displayFuncNames[x++]=  PSTR("BIG Current ");     // fuel economy fuel/100 distance units
  displayFuncNames[x++]=  PSTR("BIG Tank ");        // fuel economy distance/fuel unit
  displayFuncNames[x++]=  PSTR("BIG Tank ");        // fuel economy fuel/100 distance units
  displayFuncNames[x++]=  PSTR("Current "); 
  displayFuncNames[x++]=  PSTR("Tank "); 
  #ifdef metric
  displayFuncNames[x++]=  PSTR("ECO km/Idle li  "); 
  #else
  displayFuncNames[x++]=  PSTR("ECO mi/Idle gal "); 
  #endif
  displayFuncNames[x++]=  PSTR("CPU Monitor ");      
 
  pinMode(BrightnessPin,OUTPUT);      
  analogWrite(BrightnessPin,brightness[brightnessIdx]);      
  pinMode(EnablePin,OUTPUT);       
  pinMode(DIPin,OUTPUT);       
  pinMode(DB4Pin,OUTPUT);       
  pinMode(DB5Pin,OUTPUT);       
  pinMode(DB6Pin,OUTPUT);       
  pinMode(DB7Pin,OUTPUT);       
  delay2(500);      
 
  pinMode(ContrastPin,OUTPUT);      
  analogWrite(ContrastPin,parms[contrastIdx]);  
  LCD::init();      
  LCD::LcdCommandWrite(B00000001);  // clear display, set cursor position to zero         
  LCD::LcdCommandWrite(B10000);  // set dram to zero
  LCD::gotoXY(0,0); 
  LCD::print(getStr(PSTR("OpenGauge       ")));      
  LCD::gotoXY(0,1);      
  LCD::print(getStr(PSTR("  MPGuino v0.82m")));      

  pinMode(InjectorOpenPin, INPUT);       
  pinMode(InjectorClosedPin, INPUT);       
  pinMode(VSSPin, INPUT);            
  attachInterrupt(0, processInjOpen, FALLING);      
  attachInterrupt(1, processInjClosed, RISING);      
 
  pinMode( lbuttonPin, INPUT );       
  pinMode( mbuttonPin, INPUT );       
  pinMode( rbuttonPin, INPUT );      
 
 
  //"turn on" the internal pullup resistors      
  digitalWrite( lbuttonPin, HIGH);       
  digitalWrite( mbuttonPin, HIGH);       
  digitalWrite( rbuttonPin, HIGH);       
//  digitalWrite( VSSPin, HIGH);       
 
  //low level interrupt enable stuff      
  PCMSK1 |= (1 << PCINT8);
  enableLButton();
  enableMButton();
  enableRButton();
  PCICR |= (1 << PCIE1);       
 
  delay2(1500);       
  
  // get pre-powerdown Tank from EEPROM
  tank.loadfromEEPROM(EEPROMOFFSET_TANKTRIP);

#ifdef i2cslave
  Wire.begin(I2CADDRESS);
  Wire.onRequest(sendI2Cdata);       // slave read handler
  Wire.onReceive(receiveI2CEvent);   // slave write handler
#endif
}       
 
byte screen=0;                // display panel number - which fuel economy display to show on the LCD screen
byte holdDisplay = 0;         // set to 1 when buttons are depressed, causes fuel economy display to be suppressed during user menu input
byte alternateDisplay = 0;    // used by display functions to show alternative version (eg: L/100KM or MPG)



#define looptime 1000000ul/loopsPerSecond //1/2 second      
void loop (void){       
  if(newRun !=1)
    initGuino();//go through the initialization screen
    
  unsigned long lastActivity =microSeconds();
  unsigned long tankHold;      //state at point of last activity
 
  while(true){      
    unsigned long loopStart=microSeconds();      
    
    // LOAD CURRENT PULSES in Trip::tmpTrip INTO Trip::instant
 
    instant.reset();           //clear instant      
    cli();    // INTERRUPTS OFF, TO GRAB tmpTrip WITHOUT FEAR OF IT BEING MODIFIED
    
    instant.update(tmpTrip);   //"copy" of tmpTrip in instant now      
    tmpTrip.reset();           //reset tmpTrip first so we don't lose too many interrupts      
    instInjStart=tmpInstInjStart; 
    instInjEnd=tmpInstInjEnd; 
    instInjTot=tmpInstInjTot;     
    instInjCount=tmpInstInjCount;
    
    tmpInstInjStart=nil; 
    tmpInstInjEnd=nil; 
    tmpInstInjTot=0;     
    tmpInstInjCount=0;

    sei();  // INTERRUPTS BACK ON
    
    // TRANSMIT OUT THE SERIAL PORT TO A DATA LOGGER (PROBABLY 9600 BAUD)
    
    //send out instantmpg * 1000, instantmph * 1000, the injector/vss raw data (injector-open microseconds * 1000, injector-pulse-count * 1000, VSS-pulse-count * 1000)
    simpletx(format(instantdistanceperfuelunit()));
    simpletx(",");
    simpletx(format(instantvehiclespeed()));
    simpletx(",");
    simpletx(format(instantrpm()));
    simpletx(",");
    simpletx(format(instant.injHius*1000));
    simpletx(",");
    simpletx(format(instant.injPulses*1000));
    simpletx(",");
    simpletx(format(instant.vssPulses*1000));
    simpletx("\n");
    
    // UPDATE TRIP AND TANK OBJECTS FROM THE INSTANT OBJECT
    
    current.update(instant);   //use instant to update current      
    tank.update(instant);      //use instant to update tank      

    // TRIP TIMEOUT, A NEW TRIP STARTS, OR TRIP CONTINUES
    
    // TANK is now saved to EEPROM, each time the vehicle comes to a stop and the engine is off,
    //   or when TANK RESET or CURRENT RESET are applied.  So if the power is removed from the MPGuino,
    //   for service or whatever, TANK is loaded from EEPROM when power is re-applied.
    
//currentTripResetTimeoutUS
    if(instant.vssPulses == 0 && instant.injPulses == 0 && holdDisplay==0){
      
      // between 10 and 11 seconds of no-fuel no-VSS, save the Tank data to non-volatile EEPROM
      // so it can be loaded on the next power-up
      
      if((elapsedMicroseconds(lastActivity) > 10000000ul) && (elapsedMicroseconds(lastActivity) <= 11000000ul) && (lastActivity != nil))
        tank.savetoEEPROM(EEPROMOFFSET_TANKTRIP);
      
      if(elapsedMicroseconds(lastActivity) > parms[currentTripResetTimeoutUSIdx] && lastActivity != nil){
        analogWrite(BrightnessPin,brightness[0]);    //nitey night
        lastActivity = nil;
      }
    }else{
      
      if(lastActivity == nil){//wake up!!!
        analogWrite(BrightnessPin,brightness[brightnessIdx]);           
        lastActivity=loopStart;
        current.reset();
        tank.loopCount = tankHold;
        current.update(instant); 
        tank.update(instant); 
      }else{
        lastActivity=loopStart;
        tankHold = tank.loopCount;
      }
    }
    
    // PUSHBUTTONS FOR MENU NAVIGATION
    
    if(holdDisplay==0){
       displayFuncs[screen]();    //call the appropriate display routine      
       
    LCD::gotoXY(0,0);        
    
//see if any buttons were pressed, display a brief message if so      
    if(!(buttonState&lbuttonBit) && !(buttonState&rbuttonBit)){// left and right = initialize      
          LCD::print(getStr(PSTR("Setup ")));    
          
          // set values for math-less calibration method - which is used when refilling the tank
          unsigned long tmptankdistance = tank.distance();   // take a snapshot, in case update occurs during the updates
          unsigned long tmptankfuel = tank.fuel();
          int tmpdirty = 0;
          
          parms[calibrationDistanceIdx] = tmptankdistance;   // distance x 1000, for updating parms[vssPulsesPerDistanceUnitIdx]
          parms[calibrationFuelIdx] = tmptankfuel;           // fuel x 1000, for updating parms[microSecondsPerFuelUnitIdx]
          
          initGuino();   // edit parms[] array
          
          // math-less calibration method:
          // - if user has altered distance or fuel values to be different from tank, recalculate VSS and INJECTOR parameters
          // - this method is not as accurate as directly inputting the calibration parameters directly in the menu, if the values displayed in the config menu are of insufficient resolution

          if (parms[calibrationDistanceIdx] != tmptankdistance)
          {
              parms[vssPulsesPerDistanceUnitIdx] = recalculate(parms[vssPulsesPerDistanceUnitIdx], tmptankdistance, parms[calibrationDistanceIdx]);
              tmpdirty = 1;
          }
          
          if (parms[calibrationFuelIdx] != tmptankfuel)
          {
              parms[microSecondsPerFuelUnitIdx] = recalculate(parms[microSecondsPerFuelUnitIdx], tmptankfuel, parms[calibrationFuelIdx]);
              tmpdirty = 1;
          }

          if (tmpdirty)          
              save();    // save recalculated settings
          
          
      //}else if(!(buttonState&lbuttonBit) && !(buttonState&rbuttonBit)){// left and right = run lcd init = tank reset      
      //    LCD::print(getStr(PSTR("Init LCD "))); 
      //    LCD::init();
    }else if (!(buttonState&lbuttonBit) && !(buttonState&mbuttonBit)){// left and middle = tank reset      
          tank.reset();      
          tank.savetoEEPROM(EEPROMOFFSET_TANKTRIP);
          LCD::print(getStr(PSTR("Tank Reset ")));      
    }else if(!(buttonState&mbuttonBit) && !(buttonState&rbuttonBit)){// right and middle = current reset      
          current.reset();      
          tank.savetoEEPROM(EEPROMOFFSET_TANKTRIP);
          LCD::print(getStr(PSTR("Current Reset ")));      
    }else if(!(buttonState&lbuttonBit)){ //left is rotate through screeens to the left      
        if(screen!=0)      
          screen=(screen-1);       
        else      
          screen=displayFuncSize-1;      
        LCD::print(getStr(displayFuncNames[screen]));      
    }else if(!(buttonState&mbuttonBit)){ //middle is cycle through brightness settings      
        brightnessIdx = (brightnessIdx + 1) % brightnessLength;      
        analogWrite(BrightnessPin,brightness[brightnessIdx]);      
        LCD::print(getStr(PSTR("Brightness ")));      
        LCD::LcdDataWrite('0' + brightnessIdx);      
        LCD::print(" ");      
    }else if(!(buttonState&rbuttonBit)){//right is rotate through screeens to the left      
        screen=(screen+1)%displayFuncSize;      
        LCD::print(getStr(displayFuncNames[screen]));      
    }      
    if(buttonState!=buttonsUp)
         holdDisplay=1;
    }else{
        holdDisplay=0;
        
        //keep track of how long the loop is taking, store maximum in maxLoopLength which is elapsed microseconds
        //this value is not calculated while the config menu is operating
        unsigned long loopX=elapsedMicroseconds(loopStart);      
        if(loopX>maxLoopLength) maxLoopLength = loopX;      
    } 
    buttonState=buttonsUp;//reset the buttons      
 
    while (elapsedMicroseconds(loopStart) < (looptime));//wait for the end of the loop (1/2 second) to arrive      
  }      
 
}       
 

 




// ********* DISPLAY FUNCTIONS
 
 
// Mike's preferences:

//  Instant - L/HK, L/HOUR, SPD, RPM
//  Current - L/HK, L-USED, AVG-SPD, TRIP-DIST
//  Tank    - L/HK, L-USED, AVG-SPD, AVG-DIST
//  Total     - #TRIPS, DIST, L-USED

//  Big Real Time - L/HK (L/HOUR not moving)
//  Big Trip      - L/HK

#ifdef mike_not_ready_yet

void doDisplayInstant(){displayTripCombo(FUELCHAR,100CHAR,instantfuelperdistanceunit(100),'S',instantvehiclespeed(),FUELCHAR,HOURCHAR,instantfuelperhour(),'R',instantrpm());}      
void doDisplayCurrent(){displayTripCombo(FUELCHAR,100CHAR,current.fuelperdistanceunit(100),FUELCHAR,current.fuel(),FUELCHAR,DISTANCECHAR,instantfuelperdistanceunit(100),'R',instantrpm());}  
void doDisplayTank(){displayTripCombo(DISTANCECHAR,FUELCHAR,instantdistanceperfuelunit(),'S',instantvehiclespeed(),FUELCHAR,DISTANCECHAR,instantfuelperdistanceunit(100),'R',instantrpm());}      
void doDisplayTotals(){displayTripCombo(DISTANCECHAR,FUELCHAR,instantdistanceperfuelunit(),'S',instantvehiclespeed(),FUELCHAR,DISTANCECHAR,instantfuelperdistanceunit(100),'R',instantrpm());}      

void doDisplayBigRealTime(){
      if(instant.vssPulses == 0 && instant.injPulses == 0)
        bigNum(instantfuelperhour(),"INST",CONSUMPTIONRATESTRING);  // litres/hour whilst idling
      else
        bigNum(instant.fuelperdistanceunit(100),"INST",CONSUMPTION100STRING);  // litres/100km while in motion
}
void doDisplayBigTrip()    {bigNum(current.fuelperdistanceunit(100),"CURR",CONSUMPTION100STRING);}        // litres/100km

#endif



//075metric void doDisplayCustom(){displayTripCombo(DISTANCECHAR,FUELCHAR,instantdistanceperfuelunit(),'S',instantvehiclespeed(),FUELCHAR,'H',instantfuelperhour(),'C',current.distanceperfuelunit());}      
void doDisplayCustom(){displayTripCombo(DISTANCECHAR,FUELCHAR,instantdistanceperfuelunit(),'S',instantvehiclespeed(),FUELCHAR,DISTANCECHAR,instantfuelperdistanceunit(100),'R',instantrpm());}      

//void doDisplayCustom(){displayTripCombo('I',DISTANCECHAR,instantdistanceperfuelunit(),'S',instantfuelperhour(),'R','P',instantrpm(),'C',current.injIdleHiSec*1000);}      
//void doDisplayCustom(){displayTripCombo('I',DISTANCECHAR,995,'S',994,'R','P',999994,'C',999995);}      
void doDisplayECOIdleData(){displayTripCombo('C','E',current.ecoDistance(),FUELCHAR,current.idleFuel(),'T','E',tank.ecoDistance(),FUELCHAR,tank.idleFuel());}      
void doDisplayInstantCurrent(){displayTripCombo('I',DISTANCECHAR,instantdistanceperfuelunit(),'S',instantvehiclespeed(),'C',DISTANCECHAR,current.distanceperfuelunit(),'D',current.distance());}      
 
void doDisplayInstantTank(){displayTripCombo('I',DISTANCECHAR,instantdistanceperfuelunit(),'S',instantvehiclespeed(),'T',DISTANCECHAR,tank.distanceperfuelunit(),'D',tank.distance());}      

void doDisplayBigInstant() {                                                                 // km/litre 
  bigNum(instantdistanceperfuelunit(),"INST",EFFICIENCYSTRING);
}

void doDisplayBigCurrent() {                                                                 // km/litre
    bigNum(current.distanceperfuelunit(),"CURR",EFFICIENCYSTRING);
}

void doDisplayBigTank()    {                                                                 // km/litre
    bigNum(tank.distanceperfuelunit(),"TANK",EFFICIENCYSTRING);
}

void doDisplayBigInstantFuelPerDistanceUnit()    {                                           // litres/100km or MPG
#ifdef metric
  // alternate display every 2 seconds
  
    // 235.214 / 83000 = 0
    // (235.214 * 1000) / 86000 = 235214 / 86000 = 2.73           ... SB 27.3 ... then need to multiply by 1000 to pass result to bigNum()
    // (235.214 * 1000 * 1000) / 86000 = 235214000 / 86000 = 27350.46     ... best to make the L100KM2USMPG constant 235214000 to avoid runtime penalty multiplying by 1000 twice

  if ((alternateDisplay++ % 8) >= 4)
    bigNum((unsigned long) (L100KM2USMPG / instantfuelperdistanceunit(100)),"INST","MPG ");
  else
    bigNum(instantfuelperdistanceunit(100),"INST",CONSUMPTION100STRING);
#else
    bigNum(instantfuelperdistanceunit(100),"INST",CONSUMPTION100STRING);
#endif
}

void doDisplayBigCurrentFuelPerDistanceUnit()    {                                           // litres/100km or MPG
#ifdef metric
  // alternate display every 2 seconds
  
    // 235.214 / 83000 = 0
    // (235.214 * 1000) / 86000 = 235214 / 86000 = 2.73           ... SB 27.3 ... then need to multiply by 1000 to pass result to bigNum()
    // (235.214 * 1000 * 1000) / 86000 = 235214000 / 86000 = 27350.46     ... best to make the L100KM2USMPG constant 235214000 to avoid runtime penalty multiplying by 1000 twice

  if ((alternateDisplay++ % 8) >= 4)
    bigNum((unsigned long) (L100KM2USMPG / current.fuelperdistanceunit(100)),"CURR","MPG ");
  else
    bigNum(current.fuelperdistanceunit(100),"CURR",CONSUMPTION100STRING);
#else
    bigNum(current.fuelperdistanceunit(100),"CURR",CONSUMPTION100STRING);
#endif
}

void doDisplayBigTankFuelPerDistanceUnit()    {                                              // litres/100km or MPG
#ifdef metric
  // alternate display every 2 seconds
  if ((alternateDisplay++ % 8) >= 4)
    bigNum((unsigned long) (L100KM2USMPG / tank.fuelperdistanceunit(100)),"TANK","MPG ");
  else
    bigNum(tank.fuelperdistanceunit(100),"TANK",CONSUMPTION100STRING);
#else
    bigNum(tank.fuelperdistanceunit(100),"TANK",CONSUMPTION100STRING);
#endif
}



void doDisplayBigVSS() {
  bigNum(instantvehiclespeed(),SPEED4CHAR,"    ");
}
 
void doDisplayCurrentTripData(void){tDisplay(&current);}   //display current trip formatted data.        

void doDisplayTankTripData(void){tDisplay(&tank);}      //display tank trip formatted data.  

// display MPGuino performance statistics:
//   C%nnn Truntime              -- CPU nnn is the peak percentage of time used out of that allotted to the main loop (which is twice/second), as long as less than 100% you can add more code <g>
//   FREE MEM: nnn               -- amount of RAM left, in bytes
//
void doDisplaySystemInfo(void){      
  LCD::gotoXY(0,0);LCD::print("C%");LCD::print(format(maxLoopLength*1000/(looptime/100)));LCD::print(" T"); LCD::print(format(tank.time()));     
  unsigned long mem = memoryTest();      
  mem*=1000;      
  LCD::gotoXY(0,1);LCD::print("FREE MEM: ");LCD::print(format(mem));      
}    //display max cpu utilization and ram.        
 
void displayTripCombo(char t1, char t1L1, unsigned long t1V1, char t1L2, unsigned long t1V2,  char t2, char t2L1, unsigned long t2V1, char t2L2, unsigned long t2V2){ 
  LCD::gotoXY(0,0);LCD::LcdDataWrite(t1);LCD::LcdDataWrite(t1L1);LCD::print(format(t1V1));LCD::LcdDataWrite(' '); 
      LCD::LcdDataWrite(t1L2);LCD::print(format(t1V2)); 
  LCD::gotoXY(0,1);LCD::LcdDataWrite(t2);LCD::LcdDataWrite(t2L1);LCD::print(format(t2V1));LCD::LcdDataWrite(' '); 
      LCD::LcdDataWrite(t2L2);LCD::print(format(t2V2)); 
}      
 
//arduino doesn't do well with types defined in a script as parameters, so have to pass as void * and use -> notation.      
void tDisplay( void * r){ //display trip functions.        
  Trip *t = (Trip *)r;      
  LCD::gotoXY(0,0);LCD::print(C100STRING);LCD::print(format(t->fuelperdistanceunit(100)));LCD::print(EFFICIENCY2CHAR);LCD::print(format(t->distanceperfuelunit()));      
  LCD::gotoXY(0,1);LCD::print(DISTANCE2CHAR);LCD::print(format(t->distance()));LCD::print(FUEL2CHAR);LCD::print(format(t->fuel()));      
}      
 


// ********* LCD CLASS METHODS


//x=0..16, y= 0..1
void LCD::gotoXY(byte x, byte y){      
  byte dr=x+0x80;      
  if (y==1)       
    dr += 0x40;      
  if (y==2)       
    dr += 0x14;      
  if (y==3)       
    dr += 0x54;      
  LCD::LcdCommandWrite(dr);        
}      
 
void LCD::print(char * string){      
  byte x = 0;      
  char c = string[x];      
  while(c != 0){      
    LCD::LcdDataWrite(c);       
    x++;      
    c = string[x];      
  }      
}      
 
 
void LCD::init(){
  delay2(16);                    // wait for more than 15 msec
  pushNibble(B00110000);  // send (B0011) to DB7-4
  cmdWriteSet();
  tickleEnable();
  delay2(5);                     // wait for more than 4.1 msec
  pushNibble(B00110000);  // send (B0011) to DB7-4
  cmdWriteSet();
  tickleEnable();
  delay2(1);                     // wait for more than 100 usec
  pushNibble(B00110000);  // send (B0011) to DB7-4
  cmdWriteSet();
  tickleEnable();
  delay2(1);                     // wait for more than 100 usec
  pushNibble(B00100000);  // send (B0010) to DB7-4 for 4bit
  cmdWriteSet();
  tickleEnable();
  delay2(1);                     // wait for more than 100 usec
  // ready to use normal LcdCommandWrite() function now!
  LcdCommandWrite(B00101000);   // 4-bit interface, 2 display lines, 5x8 font
  LcdCommandWrite(B00001100);   // display control:
  LcdCommandWrite(B00000110);   // entry mode set: increment automatically, no display shift

//creating the custom fonts:
  LcdCommandWrite(B01001000);  // set cgram

 
 #if (CFG_BIGFONT_TYPE == 1)
  // default
  static byte chars[] PROGMEM ={
    B11111,B00000,B11111,B11111,B00000,
    B11111,B00000,B11111,B11111,B00000,
    B11111,B00000,B11111,B11111,B00000,
    B00000,B00000,B00000,B11111,B00000,
    B00000,B00000,B00000,B11111,B00000,
    B00000,B11111,B11111,B11111,B01110,
    B00000,B11111,B11111,B11111,B01110,
    B00000,B11111,B11111,B11111,B01110};

    for(byte x=0;x<5;x++)
      for(byte y=0;y<8;y++)
          LcdDataWrite(pgm_read_byte(&chars[y*5+x])); //write the character data to the character generator ram
          
#elif (CFG_BIGFONT_TYPE == 2)
  /* XXX: For whatever reason I can not figure out how 
   * to store more than 8 chars in the LCD CGRAM */
  static byte chars[] PROGMEM = {
    B11111, B00000, B11111, B11111, B00000, B11111, B00111, B11100, 
    B11111, B00000, B11111, B11111, B00000, B11111, B01111, B11110, 
    B00000, B00000, B00000, B11111, B00000, B11111, B11111, B11111, 
    B00000, B00000, B00000, B11111, B00000, B11111, B11111, B11111, 
    B00000, B00000, B00000, B11111, B00000, B11111, B11111, B11111, 
    B00000, B00000, B00000, B11111, B01110, B11111, B11111, B11111,
    B00000, B11111, B11111, B01111, B01110, B11110, B11111, B11111,
    B00000, B11111, B11111, B00111, B01110, B11100, B11111, B11111};
          
    /* write the character data to the character generator ram */
    for(byte x=0;x<8;x++) {
      for(byte y=0;y<8;y++) {
          LcdDataWrite(pgm_read_byte(&chars[y*8+x])); 
      }
    }
#endif


  LcdCommandWrite(B00000001);  // clear display, set cursor position to zero
  LcdCommandWrite(B10000000);  // set dram to zero

}

void  LCD::tickleEnable(){       
  // send a pulse to enable       
  digitalWrite(EnablePin,HIGH);       
  delayMicroseconds2(1);  // pause 1 ms according to datasheet       
  digitalWrite(EnablePin,LOW);       
  delayMicroseconds2(1);  // pause 1 ms according to datasheet       
}        
 
void LCD::cmdWriteSet(){       
  digitalWrite(EnablePin,LOW);       
  delayMicroseconds2(1);  // pause 1 ms according to datasheet       
  digitalWrite(DIPin,0);       
}       
 
byte LCD::pushNibble(byte value){       
  digitalWrite(DB7Pin, value & 128);       
  value <<= 1;       
  digitalWrite(DB6Pin, value & 128);       
  value <<= 1;       
  digitalWrite(DB5Pin, value & 128);       
  value <<= 1;       
  digitalWrite(DB4Pin, value & 128);       
  value <<= 1;       
  return value;      
}      
 
void LCD::LcdCommandWrite(byte value){       
  value=pushNibble(value);      
  cmdWriteSet();       
  tickleEnable();       
  value=pushNibble(value);      
  cmdWriteSet();       
  tickleEnable();       
  delay2(5);       
}       
 
void LCD::LcdDataWrite(byte value){       
  digitalWrite(DIPin, HIGH);       
  value=pushNibble(value);      
  tickleEnable();       
  value=pushNibble(value);      
  tickleEnable();       
  delay2(5);       
}       
 
 


 #if (CFG_BIGFONT_TYPE == 1)
   //32 = 0x20 = space
//   const unsigned char LcdNewChars = 5;
//   const unsigned char LcdCharHeightPix = 8;
   char bignumchars1[]={4,1,4,0, 1,4,32,0, 3,3,4,0, 1,3,4,0, 4,2,4,0,   4,3,3,0, 4,3,3,0, 1,1,4,0,   4,3,4,0, 4,3,4,0}; 
   char bignumchars2[]={4,2,4,0, 2,4,2,0,  4,2,2,0, 2,2,4,0, 32,32,4,0, 2,2,4,0, 4,2,4,0, 32,4,32,0, 4,2,4,0, 2,2,4,0};  
#elif (CFG_BIGFONT_TYPE == 2)
   //255 = 0xFF = all black character
//   const unsigned char LcdNewChars = 8;
//   const unsigned char LcdCharHeightPix = 8;
   char bignumchars1[]={7,1,8,0, 1,255,32,0,  3,3,8,0,   1,3,8,0, 255,2,255,0, 255,3,3,0, 7,3,3,0, 1,1,6,0,   7,3,8,0, 7,3,8,0};
   char bignumchars2[]={4,2,6,0, 32,255,32,0, 255,2,2,0, 2,2,6,0, 32,32,255,0, 2,2,6,0,   4,2,6,0, 32,7,32,0, 4,2,6,0, 2,2,6,0};
#endif


// displays number NN.N (fixed decimal point)
void bigNum (unsigned long t, char * txt1, char * txt2){      
//  unsigned long t = 98550ul;//number in thousandths 
//  unsigned long t = 9855ul;//number in thousandths 
//  char * txt1="INST"; 
//  char * txt2="MPG "; 
  char  dp = 32; 
 
  char * r = "009.99"; //default to 999
  if(t<=99500){ 
    r=format(t/10); //009.86 
    dp=5; 
  }else if(t<=999500){ 
    r=format(t/100); //009.86 
  }   
 
  LCD::gotoXY(0,0); 
  LCD::print(bignumchars1+(r[2]-'0')*4); 
  LCD::print(" "); 
  LCD::print(bignumchars1+(r[4]-'0')*4); 
  LCD::print(" "); 
  LCD::print(bignumchars1+(r[5]-'0')*4); 
  LCD::print(" "); 
  LCD::print(txt1); 
 
  LCD::gotoXY(0,1); 
  LCD::print(bignumchars2+(r[2]-'0')*4); 
  LCD::print(" "); 
  LCD::print(bignumchars2+(r[4]-'0')*4); 
  LCD::LcdDataWrite(dp); 
  LCD::print(bignumchars2+(r[5]-'0')*4); 
  LCD::print(" "); 
  LCD::print(txt2); 
}      

 

// ********* BASE CALCULATIONS USING VSS PULSES AND INJECTOR-OPEN TIMES 
 

unsigned long instantvehiclespeed(){      
  //unsigned long vssPulseTimeuS = (lastVSS1 + lastVSS2) / 2;
  unsigned long vssPulseTimeuS = instant.vssPulseLength/instant.vssPulses;       // microseconds between pulses
  // pulsesperhour =((1000000000/pulseintervalusec_vssPulseTimeuS)*3600)
  // speed=pulsesperhour/vssPulsesPerDistanceUnitIdx in distanceunits-per-hour
  // implemented as ...
  //   speed = ((1000000000 / vssPulsesPerDistanceUnitIdx) * 3600) / pulseintervalusec_vssPulseTimeuS
  init64(tmp1,0,1000000000ul);                           // tmp1 = 1,000,000,000
  init64(tmp2,0,parms[vssPulsesPerDistanceUnitIdx]);     // tmp2 = pulses/distance-unit
  div64(tmp1,tmp2);                                      // tmp1 = tmp1 / tmp2 = 1,000,000,000 / (pulses/distance-unit) = microseconds per distance-unit
  init64(tmp2,0,3600);                                   // tmp2 = 3600
  mul64(tmp1,tmp2);                                      // tmp1 = tmp1 / tmp2 = microseconds per distance-unit / 3600 = distance-unit per hour
  init64(tmp2,0,vssPulseTimeuS);
  div64(tmp1,tmp2);
  return tmp1[1];
}

// distanceunits_per_fuelunit in thousandths
unsigned long instantdistanceperfuelunit(){     
  unsigned long idph=instantvehiclespeed();
  unsigned long ifph=instantfuelperhour();
  if(idph == 0) return 0;
  if(ifph == 0) return 999999000;  
    // distanceunits_per_fuelunit = distance-units_per_hour / fuel-units_per_hour
    // implemented as ...           (1000 * distance-units_per_hour) / fuel-units_per_hour  
  init64(tmp1,0,1000ul);
  init64(tmp2,0,idph);
  mul64(tmp1,tmp2);
  init64(tmp2,0,ifph);
  div64(tmp1,tmp2);
  return tmp1[1];
}

// fuelunits_per_distanceunit in thousandths
unsigned long instantfuelperdistanceunit( unsigned long multiplier ){     
  unsigned long idph=instantvehiclespeed();
  unsigned long ifph=instantfuelperhour();
  if(idph == 0) return 0;
  if(ifph == 0) return 999999000;
    // fuelunits_per_distanceunit = fuel-units_per_hour / distance-units_per_hour
    // implemented as ...          (1000 * fuel-units_per_hour) / distance-units_per_hour  
  init64(tmp1,0,1000ul);
  init64(tmp2,0,ifph);
  mul64(tmp1,tmp2);
  init64(tmp2,0,idph);
  div64(tmp1,tmp2);
  
  init64(tmp2,0,multiplier);    // scale it up (eg: "1" for litres/km or "100" litres/100km
  mul64(tmp1,tmp2);
  
  return tmp1[1];
}


// for calculations involving other perhour units (eg: speed)
unsigned long instantfuelperhour(){      
//  unsigned long vssPulseTimeuS = instant.vssPulseLength/instant.vssPulses;
  
//  unsigned long instInjStart=nil; 
//unsigned long instInjEnd; 
//unsigned long instInjTot; 
  init64(tmp1,0,instInjTot);
  init64(tmp2,0,3600000000ul);
  mul64(tmp1,tmp2);
  init64(tmp2,0,1000ul);
  mul64(tmp1,tmp2);
  init64(tmp2,0,parms[microSecondsPerFuelUnitIdx]);
  div64(tmp1,tmp2);
  init64(tmp2,0,instInjEnd-instInjStart);
  div64(tmp1,tmp2);
  return tmp1[1];      
}

unsigned long instantrpm(){      
  init64(tmp1,0,instInjCount);
  init64(tmp2,0,120000000ul);
  mul64(tmp1,tmp2);
  init64(tmp2,0,1000ul);
  mul64(tmp1,tmp2);
  init64(tmp2,0,parms[injPulsesPer2Revolutions]);
  div64(tmp1,tmp2);
  init64(tmp2,0,instInjEnd-instInjStart);
  div64(tmp1,tmp2);
  return tmp1[1];      
}


// used for correcting calibration values
//   currentvalue and newvalue must already be in thousandths units (eg: 1.25 = 1250)
//   returns  calibrationvalue * currentvalue / correctvalue
unsigned long recalculate(unsigned long calibrationvalue, unsigned long currentvalue, unsigned long correctvalue)
{
  simpletx("recalc ");
  simpletx(uformat(currentvalue));     // 0000022211
  simpletx(",");
  simpletx(uformat(correctvalue));     // 0000022299
  simpletx(",");
  simpletx(uformat(calibrationvalue)); // 0000009926
  simpletx(";");
 
  init64(tmp1,0,currentvalue);
  init64(tmp2,0,1000000ul);              // multiply to obtain six decimal points of accuracy
  mul64(tmp1,tmp2);
  init64(tmp2,0,correctvalue);
  div64(tmp1,tmp2);                   // current / correct, resulting fraction in tmp1
                                      // (22211 * 1000000)/22299=996.035 = 996 (the decimal digits are lost)
  
  simpletx(uformat(tmp1[0]));
  simpletx(",");
  simpletx(uformat(tmp1[1]));
  simpletx(";");
  
  init64(tmp2,0,calibrationvalue);
  mul64(tmp1,tmp2);                   // (correction-fraction * 1000000) * original calibration value, result (new calibration value * 1000000) in tmp1
  
  simpletx(uformat(tmp1[0]));
  simpletx(",");
  simpletx(uformat(tmp1[1]));
  simpletx(";");
    
  init64(tmp2,0,1000000ul);
  div64(tmp1,tmp2);                   // back to correct magnitude
  
  simpletx(uformat(tmp1[0]));
  simpletx(", return=");
  simpletx(uformat(tmp1[1]));
  simpletx("\n");

  return tmp1[1];  
}

// ********* TRIP CLASS METHODS
 
Trip::Trip(){      
}  

// distance-unit x 1000
unsigned long Trip::distance(){      
  init64(tmp1,0,vssPulses);
  init64(tmp2,0,1000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,parms[vssPulsesPerDistanceUnitIdx]);
  div64(tmp1,tmp2);
  return tmp1[1];      
}      
 
unsigned long Trip::ecoDistance(){      
  init64(tmp1,0,vssECOPulses);
  init64(tmp2,0,1000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,parms[vssPulsesPerDistanceUnitIdx]);
  div64(tmp1,tmp2);
  return tmp1[1];      
}       
 
unsigned long Trip::vehiclespeed(){      
  if(loopCount == 0)     
     return 0;     
  init64(tmp1,0,loopsPerSecond);
  init64(tmp2,0,vssPulses);
  mul64(tmp1,tmp2);
  init64(tmp2,0,3600000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,parms[vssPulsesPerDistanceUnitIdx]);
  div64(tmp1,tmp2);
  init64(tmp2,0,loopCount);
  div64(tmp1,tmp2);
  return tmp1[1];      
}      

// fuel-units x 1000 
unsigned long  Trip::fuel(){      
  init64(tmp1,0,injHiSec);
  init64(tmp2,0,1000000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,injHius);
  add64(tmp1,tmp2);
  init64(tmp2,0,1000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,parms[microSecondsPerFuelUnitIdx]);
  div64(tmp1,tmp2);
  return tmp1[1];      
}      

unsigned long  Trip::idleFuel(){      
  init64(tmp1,0,injIdleHiSec);
  init64(tmp2,0,1000000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,injIdleHius);
  add64(tmp1,tmp2);
  init64(tmp2,0,1000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,parms[microSecondsPerFuelUnitIdx]);
  div64(tmp1,tmp2);
  return tmp1[1];      
}      

//ecoMiles
//idleGallons
 

unsigned long  Trip::distanceperfuelunit(){      
  if(vssPulses==0) return 0;      
  if(injPulses==0) return 999999000; //who doesn't like to see 999999?  :)      
    // distanceunits_per_fuelunit = distance-units_per_hour / fuel-units_per_hour
    // implemented as ...           ((usec_per_fuel-unit * 1000) * pulse_distance-units) / (inj-open-usec * distance-pulses)
  init64(tmp1,0,injHiSec);
  init64(tmp3,0,1000000);
  mul64(tmp3,tmp1);
  init64(tmp1,0,injHius);
  add64(tmp3,tmp1);
  init64(tmp1,0,parms[vssPulsesPerDistanceUnitIdx]);
  mul64(tmp3,tmp1);
 
  init64(tmp1,0,parms[microSecondsPerFuelUnitIdx]);
  init64(tmp2,0,1000);
  mul64(tmp1,tmp2);
  init64(tmp2,0,vssPulses);
  mul64(tmp1,tmp2);
 
  div64(tmp1,tmp3);
  return tmp1[1];      
} 

// fuel consumed per distance-units (litres/100km)
unsigned long  Trip::fuelperdistanceunit( unsigned long multiplier ){
  if(vssPulses==0) return 0;      
  if(injPulses==0) return 999999000; //who doesn't like to see 999999?  :)      
    // fuelunits_per_distanceunit = fuel-units_per_hour / distance-units_per_hour
    // implemented as ...           ((inj-open-usec * distance-pulses) * 1000) / (usec_per_fuel-unit * pulse_distance-units)
  init64(tmp1,0,injHiSec);
  init64(tmp3,0,1000000);
  mul64(tmp3,tmp1);
  init64(tmp1,0,injHius);
  add64(tmp3,tmp1);
  init64(tmp1,0,parms[vssPulsesPerDistanceUnitIdx]);
  mul64(tmp3,tmp1);
  init64(tmp2,0,1000);
  mul64(tmp3,tmp2);
 
  init64(tmp1,0,parms[microSecondsPerFuelUnitIdx]);
  init64(tmp2,0,vssPulses);
  mul64(tmp1,tmp2);
 
  div64(tmp3,tmp1);
  
  init64(tmp2,0,multiplier);    // scale it up (eg: "1" for litres/km or "100" litres/100km
  mul64(tmp3,tmp2);
  
  return tmp3[1];  
}

 

//return the seconds as a time mmm.ss, eventually hhh:mm too      
unsigned long Trip::time(){      
//  return seconds*1000;      
  byte d = 60;      
  unsigned long seconds = loopCount/loopsPerSecond;     
//  if(seconds/60 > 999) d = 3600; //scale up to hours.minutes if we get past 999 minutes      
  return ((seconds/d)*1000) + ((seconds%d) * 10);       
}      
 
 
void Trip::reset(){      
  loopCount=0;      
  injPulses=0;      
  injHius=0;      
  injHiSec=0;      
  vssPulses=0;  
  vssPulseLength=0;
  injIdleHiSec=0;
  injIdleHius=0;
  vssECOPulses=0;
}      
 
void Trip::update(Trip t){     
  loopCount++;  //we call update once per loop     
  vssPulses+=t.vssPulses;      
  vssPulseLength+=t.vssPulseLength;
  if(t.injPulses ==0 )  //track distance traveled with engine off
    vssECOPulses+=t.vssPulses;
  
  if(t.injPulses > 2 && t.injHius<500000){//chasing ghosts      
    injPulses+=t.injPulses;      
    injHius+=t.injHius;      
    if (injHius>=1000000){  //rollover into the injHiSec counter      
      injHiSec++;      
      injHius-=1000000;      
    }
    if(t.vssPulses == 0){    //track fuel spent sitting still
      
      injIdleHius+=t.injHius;      
      if (injIdleHius>=1000000){  //r
        injIdleHiSec++;
        injIdleHius-=1000000;      
      }      
    }
  }      
}   
 
 
void Trip::savetoEEPROM(unsigned int eepromoffset ){
  byte b = EEPROM.read(0);    // read the signature value
  
  // will save it, once Setup configuration has been updated once
  if ( b > guinosigpresaveTank) {
    ulwriteEEPROM( loopCount, eepromoffset );
    ulwriteEEPROM( injPulses, eepromoffset+4 );
    ulwriteEEPROM( injHiSec, eepromoffset+8 );
    ulwriteEEPROM( injHius, eepromoffset+12 );
    ulwriteEEPROM( injIdleHiSec, eepromoffset+16 );
    ulwriteEEPROM( injIdleHius, eepromoffset+20 );
    ulwriteEEPROM( vssPulses, eepromoffset+24 );
    ulwriteEEPROM( vssECOPulses, eepromoffset+28 );
    ulwriteEEPROM( vssPulseLength, eepromoffset+32 );
  }
}
 
 
void Trip::loadfromEEPROM(unsigned int eepromoffset ){
  byte b = EEPROM.read(0);    // read the signature value
  
  if ( b > guinosigpresaveTank) {
    loopCount = ulreadEEPROM( eepromoffset );
    injPulses =  ulreadEEPROM( eepromoffset+4 );
    injHiSec = ulreadEEPROM( eepromoffset+8 );
    injHius = ulreadEEPROM( eepromoffset+12 );
    injIdleHiSec = ulreadEEPROM( eepromoffset+16 );
    injIdleHius = ulreadEEPROM( eepromoffset+20 );
    vssPulses = ulreadEEPROM( eepromoffset+24 );
    vssECOPulses = ulreadEEPROM( eepromoffset+28 );
    vssPulseLength = ulreadEEPROM( eepromoffset+32 );
  }
  else
    reset();
}




// ********* INTEGER MATH

//the standard 64 bit math brings in  5000+ bytes
//these bring in 1214 bytes, and everything is pass by reference
unsigned long zero64[]={0,0};
 
//load most-significant-ulong and least-significant ulong, result in an[]
void init64(unsigned long  an[], unsigned long bigPart, unsigned long littlePart ){
  an[0]=bigPart;
  an[1]=littlePart;
}
 
//left shift 64 bit "number", result in an[]
void shl64(unsigned long  an[]){
 an[0] <<= 1; 
 if(an[1] & 0x80000000)
   an[0]++; 
 an[1] <<= 1; 
}
 
//right shift 64 bit "number", result in an[]
void shr64(unsigned long  an[]){
 an[1] >>= 1; 
 if(an[0] & 0x1)
   an[1]+=0x80000000; 
 an[0] >>= 1; 
}
 
//add ann to an, result in an[]
void add64(unsigned long  an[], unsigned long  ann[]){
  an[0]+=ann[0];
  if(an[1] + ann[1] < ann[1])
    an[0]++;
  an[1]+=ann[1];
}
 
//subtract ann from an, result in an[]
void sub64(unsigned long  an[], unsigned long  ann[]){
  an[0]-=ann[0];
  if(an[1] < ann[1]){
    an[0]--;
  }
  an[1]-= ann[1];
}
 
//true if an == ann
boolean eq64(unsigned long  an[], unsigned long  ann[]){
  return (an[0]==ann[0]) && (an[1]==ann[1]);
}
 
//true if an < ann
boolean lt64(unsigned long  an[], unsigned long  ann[]){
  if(an[0]>ann[0]) return false;
  return (an[0]<ann[0]) || (an[1]<ann[1]);
}
 
//divide num by den, result in num[]
void div64(unsigned long num[], unsigned long den[]){
  unsigned long quot[2];
  unsigned long qbit[2];
  unsigned long tmp[2];
  init64(quot,0,0);
  init64(qbit,0,1);
 
  if (eq64(num, zero64)) {  //numerator 0, call it 0
    init64(num,0,0);
    return;        
  }
 
  if (eq64(den, zero64)) { //numerator not zero, denominator 0, infinity in my book.
    init64(num,0xffffffff,0xffffffff);
    return;        
  }
 
  init64(tmp,0x80000000,0);
  while(lt64(den,tmp)){
    shl64(den);
    shl64(qbit);
  } 
 
  while(!eq64(qbit,zero64)){
    if(lt64(den,num) || eq64(den,num)){
      sub64(num,den);
      add64(quot,qbit);
    }
    shr64(den);
    shr64(qbit);
  }
 
  //remainder now in num, but using it to return quotient for now  
  init64(num,quot[0],quot[1]); 
}
 
 
//multiply num by den, result in an[]
void mul64(unsigned long an[], unsigned long ann[]){
  unsigned long p[2] = {0,0};
  unsigned long y[2] = {ann[0], ann[1]};
  while(!eq64(y,zero64)) {
    if(y[1] & 1) 
      add64(p,an);
    shl64(an);
    shr64(y);
  }
  init64(an,p[0],p[1]);
} 
  
  
  
// ********* EEPROM READ AND WRITE

void save(){
  EEPROM.write(EEPROMOFFSET_CONFIGURATION+0,guinosig);
  EEPROM.write(EEPROMOFFSET_CONFIGURATION+1,parmsLength);
  byte p = 0;
  for(int x=EEPROMOFFSET_CONFIGURATION+4; p < parmsLength; x+= 4){
    /*
    unsigned long v = parms[p];
    EEPROM.write(x ,(v>>24)&255);
    EEPROM.write(x + 1,(v>>16)&255);
    EEPROM.write(x + 2,(v>>8)&255);
    EEPROM.write(x + 3,(v)&255);
    */
    ulwriteEEPROM( parms[p], x );
    p++;
  }
}

byte load(){ //return 1 if loaded ok
  #ifdef usedefaults
    return 1;
  #endif
  byte b = EEPROM.read(EEPROMOFFSET_CONFIGURATION+0);
  byte c = EEPROM.read(EEPROMOFFSET_CONFIGURATION+1);
  if(b <= guinosigpremetric)
    c=9; //before added usa/metric selector

  if(b == guinosig || b == guinosigpremetric){
    byte p = 0;

    for(int x=EEPROMOFFSET_CONFIGURATION+4; p < c; x+= 4){
      /*
      unsigned long v = EEPROM.read(x);
      v = (v << 8) + EEPROM.read(x+1);
      v = (v << 8) + EEPROM.read(x+2);
      v = (v << 8) + EEPROM.read(x+3);
      parms[p]=v;
      */
      parms[p] = ulreadEEPROM(x);
      p++;
    }
    return 1;
  }
  return 0;
}


// writes to EEPROM
void ulwriteEEPROM( unsigned long ulvalue, unsigned int eepromoffset ){
    unsigned long v = ulvalue;
    EEPROM.write(eepromoffset,    (v>>24)&255);
    EEPROM.write(eepromoffset + 1,(v>>16)&255);
    EEPROM.write(eepromoffset + 2,(v>>8)&255);
    EEPROM.write(eepromoffset + 3,(v)&255);
}

// reads from EEPROM
unsigned long ulreadEEPROM( unsigned int eepromoffset ){
    unsigned long v = EEPROM.read(eepromoffset);
    v = (v << 8) + EEPROM.read(eepromoffset+1);
    v = (v << 8) + EEPROM.read(eepromoffset+2);
    v = (v << 8) + EEPROM.read(eepromoffset+3);
    return v;
}


// ********* NUMBER FORMATTING AND STRING CONVERSIONS


char fBuff[7];//used by format    

//format a number into a 5-digit decimal string, the number should already be representing thousandths
char* format(unsigned long num){
  byte dp = 3;

  while(num > 999999){
    num /= 10;
    dp++;
    if( dp == 5 ) break; // We'll lose the top numbers like an odometer
  }
  if(dp == 5) dp = 99; // We don't need a decimal point here.

// Round off the non-printed value.
  if((num % 10) > 4)
    num += 10;
  num /= 10;
  byte x = 6;
  while(x > 0){
    x--;
    if(x==dp){ //time to poke in the decimal point?{
      fBuff[x]='.';
    }else{
      fBuff[x]= '0' + (num % 10);//poke the ascii character for the digit.
      num /= 10;
    } 
  }
  fBuff[6] = 0;
  return fBuff;
}

//format a number into NNN.NN  the number should already be representing thousandths      
/*char* format(unsigned long num){      
  unsigned long d = 10000;      
  long t;      
  byte dp=3;      
  byte l=6;      
 
  if(num>9999999){      
    d=100000;      
    dp=99;      
    num/=100;      
  }else if(num>999999){      
    dp=4;      
    num/=10;      
  }      
 
  unsigned long val = num/10;      
  if ((num - (val * 10)) >= 5)  //will the first unprinted digit be greater than 4?      
    val += 1;   //round up val      
 
  for(byte x = 0; x < l; x++){      
    if(x==dp)      //time to poke in the decimal point?      
      fBuff[x]='.';      
    else{      
      t = val/d;        
      fBuff[x]= '0' + t%10;//poke the ascii character for the digit.      
      val-= t*d;      
      d/=10;            
    }      
  }      
  fBuff[6]= 0;         //good old zero terminated strings       
  return fBuff;      
} */  


char mBuff[17];//used by getStr(), uformat(), rformat()

//get a string from Flash - see PSTR() calls
char * getStr(prog_char * str){ 
  strcpy_P(mBuff, str); 
  return mBuff; 
} 

// convert long to a character string
char * uformat(unsigned long val){ 
  unsigned long d = 1000000000ul;
  for(byte p = 0; p < 10 ; p++){
    mBuff[p]='0' + (val/d);
    val=val-(val/d*d);
    d/=10;
  }
  mBuff[10]=0;
  return mBuff;
} 

// convert a character string to a long
unsigned long rformat(char * val){ 
  unsigned long d = 1000000000ul;
  unsigned long v = 0ul;
  for(byte p = 0; p < 10 ; p++){
    v=v+(d*(val[p]-'0'));
    d/=10;
  }
  return v;
} 



// ********* EDITOR TO CHANGE EACH VALUE IN THE parms[] ARRAY


void editParm(byte parmIdx){
  unsigned long v = parms[parmIdx];
  byte p=9;  //right end of 10 digit number
  //display label on top line
  //set cursor visible
  //set pos = 0
  //display v

    LCD::gotoXY(8,0);        
    LCD::print("        ");
    LCD::gotoXY(0,0);        
    LCD::print(parmLabels[parmIdx]);
    LCD::gotoXY(0,1);    
    char * fmtv=    uformat(v);
    LCD::print(fmtv);       //000000000 OK XX
    LCD::print(" OK XX");
    LCD::LcdCommandWrite(B00001110);

#ifdef cursor_first_nonzero_digit
    for(int x=9 ; x>=0 ;x--){ //do a nice thing and put the cursor at the first non zero number
      if(fmtv[x] != '0')
         p=x; 
    }
#else
    // cursor on the OK
    p=10;
#endif

  byte keyLock=1;    
  while(true){

    if(p<10)
      LCD::gotoXY(p,1);    // on a digit
    if(p==10)     
      LCD::gotoXY(11,1);   // on the OK
    if(p==11)     
      LCD::gotoXY(14,1);   // on the XX

     if(keyLock == 0){ 
        if(!(buttonState&lbuttonBit) && !(buttonState&rbuttonBit)){// left & right
            if(p<10)p=10;
            else if(p==10) p=11;
            else{
              for(int x=9 ; x>=0 ;x--){ //do a nice thing and put the cursor at the first non zero number
                if(fmtv[x] != '0')
               p=x; 
              }
            }
        }else  if(!(buttonState&lbuttonBit)){// left
            p=p-1;
            if(p==255)p=11;
        }else if(!(buttonState&rbuttonBit)){// right
             p=p+1;
            if(p==12)p=0;
        }else if(!(buttonState&mbuttonBit)){// middle
             if(p==11){  //cancel selected
                LCD::LcdCommandWrite(B00001100);
                return;
             }
             if(p==10){  //ok selected
                LCD::LcdCommandWrite(B00001100);
                parms[parmIdx]=rformat(fmtv);
                return;
             }
             
             byte n = fmtv[p]-'0';
             n++;
             if (n > 9) n=0;
             if(p==0 && n > 3) n=0;
             fmtv[p]='0'+ n;
             LCD::gotoXY(0,1);        
             LCD::print(fmtv);
             LCD::gotoXY(p,1);        
             if(parmIdx==contrastIdx)//adjust contrast dynamically
                 analogWrite(ContrastPin,rformat(fmtv));  


        }

      if(buttonState!=buttonsUp)
         keyLock=1;
     }else{
        keyLock=0;
     }
      buttonState=buttonsUp;
      delay2(125);
  }      
  
}

void initGuino(){ //edit all the parameters
  for(int x = 0;x<parmsLength;x++)
    editParm(x);
  save();
  holdDisplay=1;
}  



// ********* MORE TIMER AND INITIALIZATION ROUTINES


unsigned long millis2(){
	return timer2_overflow_count * 64UL * 2 / (16000000UL / 128000UL);
}

void delay2(unsigned long ms){
	unsigned long start = millis2();
	while (millis2() - start < ms);
}

/* Delay for the given number of microseconds.  Assumes a 16 MHz clock. 
 * Disables interrupts, which will disrupt the millis2() function if used
 * too frequently. */
void delayMicroseconds2(unsigned int us){
	uint8_t oldSREG;
	if (--us == 0)	return;
	us <<= 2;
	us -= 2;
	oldSREG = SREG;
	cli();
	// busy wait
	__asm__ __volatile__ (
		"1: sbiw %0,1" "\n\t" // 2 cycles
		"brne 1b" : "=w" (us) : "0" (us) // 2 cycles
	);
	// reenable interrupts.
	SREG = oldSREG;
}

void init2(){
	// this needs to be called before setup() or some functions won't
	// work there
	sei();
	
	// timer 0 is used for millis2() and delay2()
	timer2_overflow_count = 0;
	// on the ATmega168, timer 0 is also used for fast hardware pwm
	// (using phase-correct PWM would mean that timer 0 overflowed half as often
	// resulting in different millis2() behavior on the ATmega8 and ATmega168)
        TCCR2A=1<<WGM20|1<<WGM21;
	// set timer 2 prescale factor to 64
        TCCR2B=1<<CS22;


//      TCCR2A=TCCR0A;
//      TCCR2B=TCCR0B;
	// enable timer 2 overflow interrupt
	TIMSK2|=1<<TOIE2;
	// disable timer 0 overflow interrupt
	TIMSK0&=!(1<<TOIE0);
}


// transmit out the serial port at 9600 baud
#define myubbr (16000000/16/9600-1)
void simpletx( char * string ){
 if (UCSR0B != (1<<TXEN0)){ //do we need to init the uart?
    UBRR0H = (unsigned char)(myubbr>>8);
    UBRR0L = (unsigned char)myubbr;
    UCSR0B = (1<<TXEN0);//Enable transmitter
    UCSR0C = (3<<UCSZ00);//N81
 }
 while (*string)
 {
   while ( !( UCSR0A & (1<<UDRE0)) );
   UDR0 = *string++; //send the data
 }
}



 
// return the number of bytes currently free in RAM      
extern int  __bss_end; 
extern int  *__brkval; 
int memoryTest(){ 
  int free_memory; 
  if((int)__brkval == 0) 
    free_memory = ((int)&free_memory) - ((int)&__bss_end); 
  else 
    free_memory = ((int)&free_memory) - ((int)__brkval); 
  return free_memory; 
} 
